import time
import pyvisa
from threading import Lock
from pyvisa.errors import VisaIOError
import pymodaq.daq_utils.daq_utils as utils
from pymodaq.utils.logger import set_logger, get_module_name

logger = set_logger(get_module_name(__file__), add_to_console=False)
visa_rm = pyvisa.ResourceManager()

_infos = visa_rm.list_resources_info()
COMPORTS = []
for k in _infos.keys():
    COMPORTS.append(_infos[k].alias)

lock = Lock()


class AgilisChannelError(Exception):
    pass


class AgilisAxisError(Exception):
    pass


class AgilisSerial:
    channel_indexes = [1, 2, 3, 4]  # for 'AG-UC8' else [1, 2]
    axis_indexes = [1, 2]
    _steps = {axis: 0 for axis in axis_indexes}

    def __init__(self):
        self._controller = None
        self._info = None
        self._timeout_wait_isready_ms = 10000

    def init_com_remote(self, com_port):
        self.open(com_port)
        self.reset()
        time.sleep(1)
        info = self.get_infos()
        self.set_local_remote('remote')
        return info

    def open(self, com_port):
        if com_port in COMPORTS:
            self._controller = visa_rm.open_resource(com_port, baud_rate=921600)
            time.sleep(1)

            self._controller.read_termination = self._controller.CR + self._controller.LF
            self._controller.write_termination = self._controller.CR + self._controller.LF
            self._controller.timeout = 10

    def get_infos(self):
        if self._controller is not None:
            if self._info is None:
                self._info = self.query('VE')
        if 'AG-UC8' in self._info:
            self.channel_indexes = [1, 2, 3, 4]
        else:
            self.channel_indexes = [1, 2]
        return self._info

    def set_local_remote(self, mode='remote'):
        if mode == 'remote':
            self.write('MR')
        else:
            self.write('ML')

    def reset(self):
        self.write('RS')

    def stop(self, axis: int):
        command = f'{axis:.0f}ST'
        self.write(command)

    def select_channel(self, channel_index: int):
        if channel_index not in self.channel_indexes:
            raise AgilisChannelError(f'The specified channel ({channel_index}) is not available in {self.channel_indexes}')
        order = "CC" + str(channel_index)
        self.write(order)

    def get_channel(self):
        channel = self.query('CC?')
        return int(channel[2:])

    def check_axis_index(self, axis_index: int):
        if axis_index not in self.axis_indexes:
            raise AgilisAxisError(
                f'The specified axis ({axis_index}) is not available in {self.axis_index}')

    def get_axis_isready(self, axis):
        self.check_axis_index(axis)
        command = f'{axis:.0f}TS'
        status = self.query(command)
        return status == f'{command}0'

    def wait_axis_ready(self, axis):
        time_start = time.perf_counter()
        while not self.get_axis_isready(axis):
            time.sleep(0.05)
            if time.perf_counter() - time_start > self._timeout_wait_isready_ms / 1000:
                self.stop(axis)
                raise TimeoutError(f"axis {axis} could'nt be ready after an elapsed time of"
                                   f" {self._timeout_wait_isready_ms} ms")

    def wait_query_is_not_none(self, axis):
        time_start = time.perf_counter()
        while not self.get_axis_isready(axis):
            time.sleep(0.05)
            if time.perf_counter() - time_start > self._timeout_wait_isready_ms / 1000:
                self.stop(axis)
                raise TimeoutError(f"axis {axis} could'nt be ready after an elapsed time of"
                                   f" {self._timeout_wait_isready_ms} ms")

    def move_rel(self, axis: int, steps: int):
        self.check_axis_index(axis)
        order = f'{axis:.0f}PR{steps:.0f}'
        self.write(order)
        self._steps[axis] += steps

    def counter_to_zero(self, axis):
        self.check_axis_index(axis)
        command = f'{axis:.0f}ZP'
        self.write(command)

    @utils.timer
    def get_step_counter(self, axis, read_controller=True):
        """
        Returns the number of accumulated steps in forward direction minus the number of steps in backward direction
        since powering the controller or since the last ZP (zero position) command
        """

        self.check_axis_index(axis)
        if read_controller:
            self.wait_axis_ready(axis)
            command = f'{axis:.0f}TP'
            steps_string = self.query(command)
            if steps_string is None or command not in steps_string:
                steps = self._steps[axis]
            else:
                steps = int(steps_string.split(command)[1])
                self._steps[axis] = steps
        else:
            steps = self._steps[axis]
        return steps

    def is_at_limits(self):
        """
        check if both axis of current channel are at the limit (if any)
        Returns
        -------
        bool: status of axis 1
        bool: status of axis 2
        """
        command = f'PH'
        ret = self.query(command)
        if ret == 'PH0':
            return False, False
        elif ret == 'PH1':
            return True, False
        elif ret == 'PH2':
            return False, True
        elif ret == 'PH3':
            return True, True

    def close(self):
        self._controller.close()

    def query(self, command: str):
        value = None
        try:
            lock.acquire()
            time_start = time.perf_counter()
            while value is None:
                self.write(command, isquery=True)
                value = self.flush_read()
                ret = self.check_errors(command)
                logger.debug(f'Error code {ret} returned from the query of the write of {command}')
                if value is None:
                    time.sleep(0.05)
                    if time.perf_counter() - time_start > self._timeout_wait_isready_ms / 1000:
                        raise TimeoutError(f"Timeout append during query of command {command}")
        except VisaIOError as e:
            logger.debug(str(e))
        finally:
            lock.release()
        return value

    def check_errors(self, command=''):
        ret = self._controller.query('TE')
        if ret != 'TE0':
            logger.warning(f'Error code {ret} returned from the query of the command {command}')
        return ret

    def write(self, command: str, isquery=True):
        try:
            if not isquery:
                lock.acquire()
            self._controller.write(command)
            if not isquery:
                ret = self.check_errors(command)
                logger.debug(f'Error code {ret} returned from the query of the write of {command}')
        except VisaIOError as e:
            logger.debug(str(e))
        finally:
            if not isquery:
                lock.release()

    def flush_read(self):
        ret = None
        while True:
            try:
                ret = self._controller.read()
                logger.debug(f'Read buffer was {ret}')
            except pyvisa.errors.VisaIOError:
                #  expected timeout
                break
        return ret


if __name__ == '__main__':
    ag = AgilisSerial()
    info = ag.init_com_remote('COM9')

    ag.close()



# -*- coding: utf-8 -*-
"""
Created on Mon Jan  9 15:41:28 2023

@author: weber


Wrapper around the serial communication with a particular SMC100 controller whose most methods from the SerialBase object 

"""

import pyvisa
import numpy as np
from pymodaq_plugins_newport.hardware.serial_base import SerialBase


class SMC100(SerialBase):
   
    def init_communication(self, com_port, axis=1):
        if com_port in self.com_ports:
            super().init_communication(com_port, axis)
            self._controller.baud_rate = 57600
        else:
            raise IOError('{:s} is not a valid port'.format(com_port))

    def _str_to_float(self, command:str, string:str) -> float:
        return float(string.split(f'{command}')[1][:-2])


    def get_position(self, axis=1):
        """ return the given axis position always in mm
        """
        command = f'{axis}TP'
        self._write_command(command)
        pos = self._str_to_float(command, self.read())
        return pos
    
    def get_velocity(self, axis=1):
        command = f'{axis}VA?'
        self._write_command(command)
        pos = self._str_to_float(command[:-1],  self.read())
        return pos
    
    def get_velocity_max(self, axis=1):
        raise NotImplementedError
    
    def move_home(self, axis=1):
        self._write_command(f'{axis}OR')
        
    
    
    
if __name__ == '__main__':
    controller = SMC100()
    controller.init_communication('COM5')
    try:
        print(controller.get_controller_infos())
        print(f'Position is : {controller.get_position()}')
        controller.move_axis('REL', pos=-1.)
    except Exception as e:
        print(str(e))
    finally:
        controller.close_communication()
        
        
    
    


# -*- coding: utf-8 -*-
"""
Created on Mon Jan  9 15:57:26 2023

@author: lb19g16
"""
from pymodaq.control_modules.move_utility_classes import DAQ_Move_base, main  # base class
from pymodaq.control_modules.move_utility_classes import comon_parameters_fun  # common set of parameters for all actuators
from pymodaq.daq_utils.daq_utils import ThreadCommand, getLineInfo  # object used to send info back to the main thread
from easydict import EasyDict as edict  # type of dict

from pymodaq_plugins_newport.hardware.smc100 import SMC100
import pyvisa

VISA_rm = pyvisa.ResourceManager()
infos = VISA_rm.list_resources_info()
com_ports = [infos[key].alias for key in infos.keys() if infos[key].alias is not None]
VISA_rm.close()


class DAQ_Move_Newport_SMC100(DAQ_Move_base):
    """Plugin for the Template Instrument

    This object inherits all functionality to communicate with PyMoDAQ Module through inheritance via DAQ_Move_base
    It then implements the particular communication with the instrument

    Attributes:
    -----------
    controller: object
        The particular object that allow the communication with the hardware, in general a python wrapper around the
         hardware library
    # TODO add your particular attributes here if any

    """
    _controller_units = 'mm'  # TODO for your plugin: put the correct unit here
    is_multiaxes = True  # TODO for your plugin set to True if this plugin is controlled for a multiaxis controller
    axes_names = ['1']  # The axis list represents the number of smc controllers, indexed: first=1, second=2 etc.
    _epsilon = 0.0001
    params = [  {'title': 'COM Port:', 'name': 'com_port', 'type': 'list', 'limits': com_ports, 'value': 'COM17'},
              # TODO for your custom plugin: elements to be added here as dicts in order to control your custom stage
                ] + comon_parameters_fun(is_multiaxes, axes_names)

    def ini_attributes(self):
        #  TODO declare the type of the wrapper (and assign it to self.controller) you're going to use for easy
        #  autocompletion
        self.controller: SMC100 = None

        #TODO declare here attributes you want/need to init with a default value
        pass

    def get_actuator_value(self):
        """Get the current value from the hardware with scaling conversion.

        Returns
        -------
        float: The position obtained after scaling conversion.
        """

        axis = int(self.settings.child('multiaxes', 'axis').value())
        pos = self.controller.get_position(axis)  # when writing your own plugin replace this line
        print(f'pos is {pos}')
        pos = self.get_position_with_scaling(pos)
        return pos

    def close(self):
        """Terminate the communication protocol"""
        axis = int(self.settings.child('multiaxes', 'axis').value())
        self.controller.close_communication(axis)  # when writing your own plugin replace this line

    def commit_settings(self, param):
        """Apply the consequences of a change of value in the detector settings

        Parameters
        ----------
        param: Parameter
            A given parameter (within detector_settings) whose value has been changed by the user
        """
        ## TODO for your custom plugin
        if param.name() == "a_parameter_you've_added_in_self.params":
           self.controller.your_method_to_apply_this_param_change()
        else:
            pass

    def ini_stage(self, controller=None):
        """Actuator communication initialization

        Parameters
        ----------
        controller: (object)
            custom object of a PyMoDAQ plugin (Slave case). None if only one actuator by controller (Master case)

        Returns
        -------
        info: str
        initialized: bool
            False if initialization failed otherwise True
        """

        self.ini_stage_init(old_controller=controller,
                            new_controller=SMC100())
        if self.settings['multiaxes', 'multi_status'] == "Master":
            self.controller.init_communication(
                self.settings['com_port'])
        axis = int(self.settings.child('multiaxes', 'axis').value())
        info = self.controller.get_controller_infos(axis)
        initialized = True  # todo
        return info, initialized

    def move_abs(self, value):
        """ Move the actuator to the absolute target defined by value

        Parameters
        ----------
        value: (float) value of the absolute target positioning
        """

        value = self.check_bound(value)  #if user checked bounds, the defined bounds are applied here
        self.target_value = value
        value = self.set_position_with_scaling(value)  # apply scaling if the user specified one
        ## TODO for your custom plugin
        axis = int(self.settings.child('multiaxes', 'axis').value())
        self.controller.move_axis(axis=axis,pos=value)  # when writing your own plugin replace this line

    def move_rel(self, value):
        """ Move the actuator to the relative target actuator value defined by value

        Parameters
        ----------
        value: (float) value of the relative target positioning
        """
        value = self.check_bound(self.current_position + value) - self.current_position
        self.target_value = value + self.current_position
        value = self.set_position_relative_with_scaling(value)

        axis = int(self.settings.child('multiaxes', 'axis').value())
        self.controller.move_axis('REL',axis=axis,pos=value)  # when writing your own plugin replace this line


    def move_home(self):
        """Call the reference method of the controller"""

        ## TODO for your custom plugin
        axis = int(self.settings.child('multiaxes', 'axis').value())
        self.controller.move_home(axis)  # when writing your own plugin replace this line


    def stop_motion(self):
      """Stop the actuator and emits move_done signal"""

      axis = int(self.settings.child('multiaxes', 'axis').value())
      self.controller.stop_motion(axis)  # when writing your own plugin replace this line


if __name__ == '__main__':
    main(__file__)